#!/usr/bin/env python
"""
_LoadRunning_

Oracle implementation for loading bl_runjob records for active jobs
"""


from WMCore.BossAir.MySQL.LoadRunning import LoadRunning as MySQLLoadRunning

class LoadRunning(MySQLLoadRunning):
    """
    _LoadRunning_

    """
