#!/usr/bin/env python
"""
_LoadDBSFilesByDAS_

Oracle implementation of DBS3Buffer.LoadDBSFilesByDAS
"""

from WMComponent.DBS3Buffer.MySQL.LoadDBSFilesByDAS import LoadDBSFilesByDAS as MySQLLoadDBSFilesByDAS

class LoadDBSFilesByDAS(MySQLLoadDBSFilesByDAS):
    pass
