#!/usr/bin/env python
"""
_DeleteCheckWorkflow_

Oracle implementation of DeleteCheckWorkflow

"""
__all__ = []



from WMCore.WMBS.MySQL.Workflow.DeleteCheck import DeleteCheck as MySQLDeleteCheck

class DeleteCheck(MySQLDeleteCheck):
    """
    Same as MySQL version

    """
