#!/usr/bin/env python
"""
_GetChildIDsByID_

Oracle implementation of ChildIDsByID
"""




from WMCore.WMBS.MySQL.Files.GetChildIDsByID import GetChildIDsByID \
     as GetChildIDsMySQL

class GetChildIDsByID(GetChildIDsMySQL):
    pass
